local hcTk = HCToolkit
local colors = hcTk.Constants.Colors

local GetRaidDifficultyID = GetRaidDifficultyID
local UnitIsGroupLeader = UnitIsGroupLeader
local WrapTextInColorCode = WrapTextInColorCode

function hcTk.AvatarEncounterStart()
    hcTk.Encounter["DarkMarks"] = {}
    hcTk.Encounter["DarkMarksWipeScheduled"] = false
    hcTk.Encounter["TaintedSoakLeft"] = {}
    hcTk.Encounter["TaintedSoakRight"] = {}
    for k, v in pairs(hcTk.Group) do
        v["DarkMarkExpiration"] = 0
        v["TaintedExpiration"] = 0
        v["TaintedGroup"] = ""
        v["TaintedStacks"] = 0
        v["TaintedSoaking"] = false
    end

    C_Timer.After(0.5, hcTk.AvatarSendTaintedSoakGroups)
end

function hcTk.AvatarEncounterEnd()
    hcTk.Encounter["DarkMarks"] = nil
    hcTk.Encounter["DarkMarksWipeScheduled"] = nil
    hcTk.Encounter["TaintedSoakLeft"] = nil
    hcTk.Encounter["TaintedSoakRight"] = nil
    for k, v in pairs(hcTk.Group) do
        v["DarkMarkExpiration"] = nil
        v["TaintedExpiration"] = nil
        v["TaintedGroup"] = nil
        v["TaintedStacks"] = nil
        v["TaintedSoaking"] = nil
    end
end

function hcTk.AvatarCombatLog(...)
    local msg = select(2, ...)
    if (hcTk.AvatarCombatLogFns[msg] ~= nil) then
        hcTk.AvatarCombatLogFns[msg](...)
    end
end

function hcTk.AvatarSpellAuraApplied(...)
    local destGUID, _, _, _, spellId = select(8, ...)
    if (spellId == 239739) then -- Dark Mark
        hcTk.AvatarDarkMarkHandler(destGUID)
    elseif (spellId == 242017) then -- Black Winds
        hcTk.AvatarBlackWindsHandler(destGUID)
    elseif (spellId == 240746) then -- Tainted Matrix (Beam)
        hcTk.AvatarTaintedMatrixAppliedHandler(destGUID)
    elseif (spellId == 240728) then -- Tainted Essence (Stacking Debuff)
        hcTk.AvatarTaintedEssenceAppliedHandler(destGUID, 1)
    end
end

function hcTk.AvatarSpellAuraAppliedDose(...)
    local destGUID, _, _, _, spellId, _, _, _, stacks = select(8, ...)
    if (spellId == 240728) then -- Tainted Essence (Stacking Debuff)
        hcTk.AvatarTaintedEssenceAppliedHandler(destGUID, stacks)
    end
end

function hcTk.AvatarSpellAuraRemoved(...)
    local destGUID, _, _, _, spellId = select(8, ...)
    if (spellId == 240746) then -- Tainted Matrix (Beam)
        hcTk.AvatarTaintedMatrixRemovedHandler(destGUID)
    elseif (spellId == 240728) then -- Tainted Essence (Stacking Debuff)
        hcTk.AvatarTaintedEssenceRemovedHandler(destGUID)
    end
end

function hcTk.AvatarSpellCastSuccess(...)
    local destGUID, _, _, _, spellId = select(8, ...)
    if (spellId == 235597) then -- Annihilation (Transition)
        hcTk.AvatarWipeTainted()
        hcTk.AvatarAvalanchReminder()
    end
end

function hcTk.AvatarUnitDied(...)
    
end

function hcTk.AvatarBlackWindsHandler(destGUID)
    if (destGUID == hcTk.PlayerGUID and hcTk.GenericGet("MattySmite")) then
        PlaySoundFile("Interface\\AddOns\\HCToolkit\\Media\\ThatsWrong.mp3", "Master")
    end
end

function hcTk.AvatarDarkMarkHandler(destGUID)
    hcTk.Encounter["DarkMarks"][#hcTk.Encounter["DarkMarks"] + 1] = destGUID
    hcTk.Group[destGUID]["DarkMarkExpiration"] = GetTime() + 6 + (2 * (#hcTk.Encounter["DarkMarks"] - 1)) -- Mark #1: 6s, Mark #2: 8s, Mark #3: 10s

    if (#hcTk.Encounter["DarkMarks"] == 3 and hcTk.GenericGetBossOption("TOS", "Avatar", "DarkMarksGroup")) then -- Dark Marks Complete, Trigger Printing
        HCToolkitGroupFrame:SetScript("OnUpdate", hcTk.AvatarPrintGroupDarkMarks)
    end

    if (hcTk.PlayerGUID == destGUID and hcTk.GenericGetBossOption("TOS", "Avatar", "DarkMarkPersonal")) then
        HCToolkitPersonalFrame:SetScript("OnUpdate", hcTk.AvatarPrintPersonalDarkMark)
    end

    if (not hcTk.Encounter["DarkMarksWipeScheduled"]) then
        hcTk.Encounter["DarkMarksWipeScheduled"] = true
        C_Timer.After(11, hcTk.AvatarWipeDarkMarks)
    end
end

function hcTk.AvatarPrintPersonalDarkMark()
    local positions = { [1] = "Left", [2] = "Middle", [3] = "Right" }
    local markers = { [1] = 3, [2] = 4, [3] = 5 }

    for k, m in ipairs(hcTk.Encounter["DarkMarks"]) do
        if (hcTk.PlayerGUID == m) then
            if (GetTime() > hcTk.Group[m]["DarkMarkExpiration"]) then
                HCToolkitPersonalFrame:SetScript("OnUpdate", nil)
                hcTk.FrameUpdateText(HCToolkitPersonalFrame, "")
            else
                hcTk.FrameUpdateText(HCToolkitPersonalFrame, string.format("%s%s: %s - %.1f", hcTk.MarkerIdToString(markers[k]), positions[k], hcTk.UnitPlayerString(hcTk.Group[m]), math.max(0, hcTk.Group[m]["DarkMarkExpiration"] - GetTime())))
            end
        end
    end
end

function hcTk.AvatarPrintGroupDarkMarks()
    local positions = { [1] = "Left", [2] = "Middle", [3] = "Right" }
    local markers = { [1] = 3, [2] = 4, [3] = 5 }

    local str = ""
    for k, m in ipairs(hcTk.Encounter["DarkMarks"]) do
        str = str .. string.format("%s%s: %s - %.1f\n", hcTk.MarkerIdToString(markers[k]), positions[k], hcTk.UnitPlayerString(hcTk.Group[m]), math.max(0, hcTk.Group[m]["DarkMarkExpiration"] - GetTime()))
    end
    
    if (GetTime() > hcTk.Group[hcTk.Encounter["DarkMarks"][#hcTk.Encounter["DarkMarks"]]]["DarkMarkExpiration"]) then
        HCToolkitGroupFrame:SetScript("OnUpdate", nil)
        hcTk.FrameUpdateText(HCToolkitGroupFrame, "")
    else
        hcTk.FrameUpdateText(HCToolkitGroupFrame, str)
    end
end

function hcTk.AvatarWipeDarkMarks()
    hcTk.Encounter["DarkMarks"] = {}
    for k, v in pairs(hcTk.Group) do
        v["DarkMarkExpiration"] = 0
    end
    hcTk.Encounter["DarkMarksWipeScheduled"] = false
end

function hcTk.AvatarTaintedMatrixAppliedHandler(destGUID)
    if (hcTk.Group[destGUID] ~= nil) then
        hcTk.Group[destGUID]["TaintedSoaking"] = true
    end
end

function hcTk.AvatarTaintedMatrixRemovedHandler(destGUID)
    if (hcTk.Group[destGUID] ~= nil) then
        hcTk.Group[destGUID]["TaintedSoaking"] = false
    end
end

function hcTk.AvatarTaintedEssenceAppliedHandler(destGUID, stacks)
    if (hcTk.Group[destGUID] ~= nil) then
        hcTk.Group[destGUID]["TaintedStacks"] = stacks
        hcTk.Group[destGUID]["TaintedExpiration"] = GetTime() + 60
        hcTk.AvatarTaintedEssenceSoakAlert(destGUID)
    end
end

function hcTk.AvatarTaintedEssenceRemovedHandler(destGUID)
    if (hcTk.Group[destGUID] ~= nil) then
        hcTk.Group[destGUID]["TaintedStacks"] = 0
        hcTk.Group[destGUID]["TaintedExpiration"] = 0
    end
end

function hcTk.AvatarTaintedEssenceIsPlayerInGroup(destGUID, group)
    for k, guid in ipairs(group) do
        if (guid == destGUID) then return k end
    end

    return nil
end

function hcTk.AvatarTaintedEssenceShouldAlert(destGUID, group)
    local playerIndex = hcTk.AvatarTaintedEssenceIsPlayerInGroup(hcTk.PlayerGUID, group)
    local soakerIndex = hcTk.AvatarTaintedEssenceIsPlayerInGroup(destGUID, group)
    if (playerIndex == nil or soakerIndex == nil) then return false end
    return (playerIndex ~= soakerIndex) and ((#group == soakerIndex and playerIndex == 1) or (soakerIndex + 1 == playerIndex))
end

function hcTk.AvatarTaintedEssenceSoakAlert(destGUID)
    if (hcTk.GenericGetBossOption("TOS", "Avatar", "TaintedSoakPersonal") and hcTk.Group[hcTk.PlayerGUID]["TaintedGroup"] ~= "" 
        and hcTk.Group[destGUID] ~= nil and hcTk.Group[hcTk.PlayerGUID] ~= nil 
        and hcTk.Encounter["TaintedSoakLeft"] ~= nil and hcTk.Encounter["TaintedSoakRight"] ~= nil) then

        hcTk.DebugPrint("Avatar Soak Alert!!")
        if (hcTk.AvatarTaintedEssenceShouldAlert(destGUID, hcTk.Encounter["TaintedSoakLeft"])) then
            hcTk.FrameUpdateText(HCToolkitPersonalFrame, hcTk.AvatarTaintedEssenceSoakAlertText(hcTk.Group[destGUID]))
        elseif (hcTk.Group[hcTk.PlayerGUID]["TaintedGroup"] == "Left" and hcTk.Group[destGUID]["TaintedGroup"] == "Left") then
            hcTk.FrameUpdateText(HCToolkitPersonalFrame, "")
        elseif (hcTk.AvatarTaintedEssenceShouldAlert(destGUID, hcTk.Encounter["TaintedSoakRight"])) then
            hcTk.FrameUpdateText(HCToolkitPersonalFrame, hcTk.AvatarTaintedEssenceSoakAlertText(hcTk.Group[destGUID]))
        elseif (hcTk.Group[hcTk.PlayerGUID]["TaintedGroup"] == "Right" and hcTk.Group[destGUID]["TaintedGroup"] == "Right") then
            hcTk.FrameUpdateText(HCToolkitPersonalFrame, "")
        end
    end
end

function hcTk.AvatarTaintedEssenceSoakAlertText(player)
    return string.format("%s %s - %s", "Soaking Next!", hcTk.UnitPlayerString(player), hcTk.AvatarTaintedEssenceSoakPrintStacks(player["TaintedStacks"]))
end

function hcTk.AvatarTaintedEssenceSoakPrintPlayer(player)
    if (not player["alive"]) then
        return string.format("%s %s %s: %s | %ss\n", hcTk.MarkerIdToString(8), hcTk.UnitPlayerString(player), hcTk.MarkerIdToString(8), hcTk.AvatarTaintedEssenceSoakPrintStacks(player["TaintedStacks"]), hcTk.AvatarTaintedEssenceSoakPrintDuration(player["TaintedExpiration"]))
    elseif (player["TaintedSoaking"]) then
        return string.format("%s %s: %s | %ss\n", WrapTextInColorCode(">", colors["lightPurple"]), hcTk.UnitPlayerString(player), hcTk.AvatarTaintedEssenceSoakPrintStacks(player["TaintedStacks"]), hcTk.AvatarTaintedEssenceSoakPrintDuration(player["TaintedExpiration"]))
    else
        return string.format("%s: %s | %ss\n", hcTk.UnitPlayerString(player), hcTk.AvatarTaintedEssenceSoakPrintStacks(player["TaintedStacks"]), hcTk.AvatarTaintedEssenceSoakPrintDuration(player["TaintedExpiration"]))
    end
end

function hcTk.AvatarTaintedEssenceSoakPrintDuration(expirationTime)
    if (expirationTime < GetTime()) then
        return "0"
    else
        return string.format("%d", math.abs(GetTime() - expirationTime))
    end
end

function hcTk.AvatarTaintedEssenceSoakPrintStacks(stacks)
    if (stacks >= 7) then
        return WrapTextInColorCode(string.format("%d", stacks), colors["red"])
    elseif (stacks >= 5) then
        return WrapTextInColorCode(string.format("%d", stacks), colors["gold"])
    elseif (stacks >= 1) then
        return WrapTextInColorCode(string.format("%d", stacks), colors["lightGreen"])
    else
        return WrapTextInColorCode(string.format("%d", stacks), colors["white"])
    end
end

function hcTk.AvatarTaintedEssenceSoakPrintGroup(group)
    local str = ""

    for k, guid in ipairs(group) do
        str = str .. hcTk.AvatarTaintedEssenceSoakPrintPlayer(hcTk.Group[guid])
    end

    if (string.len(str) >= 2) then
        str = str:sub(1, -2)
    end

    return str
end

function hcTk.AvatarTaintedEssenceSoakPrintGroups()
    local leftGroup = ""
    local rightGroup = ""
    local merger = ""

    if (hcTk.GenericGetBossOption("TOS", "Avatar", "TaintedSoakGroup") == "BOTH" or hcTk.Group[hcTk.PlayerGUID]["TaintedGroup"] == "Left") then
        leftGroup = hcTk.AvatarTaintedEssenceSoakPrintGroup(hcTk.Encounter["TaintedSoakLeft"])
    end

    if (hcTk.GenericGetBossOption("TOS", "Avatar", "TaintedSoakGroup") == "BOTH" or hcTk.Group[hcTk.PlayerGUID]["TaintedGroup"] == "Right") then
        rightGroup = hcTk.AvatarTaintedEssenceSoakPrintGroup(hcTk.Encounter["TaintedSoakRight"])
    end

    if (leftGroup ~= "" and rightGroup ~= "") then
        merger = "\n\n"
    end

    hcTk.FrameUpdateText(HCToolkitGroupFrame, string.format("%s%s%s", leftGroup, merger, rightGroup))
end

function hcTk.AvatarReceiveTaintedSoakGroups(leftGroup, rightGroup)
    hcTk.Encounter["TaintedSoakLeft"] = leftGroup
    hcTk.Encounter["TaintedSoakRight"] = rightGroup
    if (hcTk.AvatarTaintedEssenceIsPlayerInGroup(hcTk.PlayerGUID, leftGroup) ~= nil) then
        hcTk.Group[hcTk.PlayerGUID]["TaintedGroup"] = "Left"
    elseif (hcTk.AvatarTaintedEssenceIsPlayerInGroup(hcTk.PlayerGUID, rightGroup) ~= nil) then
        hcTk.Group[hcTk.PlayerGUID]["TaintedGroup"] = "Right"
    end

    if (hcTk.GenericGetBossOption("TOS", "Avatar", "TaintedSoakGroup") ~= "NONE") then
        HCToolkitGroupFrame:SetScript("OnUpdate", hcTk.AvatarTaintedEssenceSoakPrintGroups)
    end
end

function hcTk.AvatarSendTaintedSoakGroups()
    if (UnitIsGroupLeader(hcTk.Group[hcTk.PlayerGUID]["unit"]) and GetRaidDifficultyID() == hcTk.Constants.RaidDifficulties["Mythic"]) then
        local data = {
            ["TaintedSoakLeft"] = hcTk.GroupNameToGUIDs(hcTk.DB.profile.Assignments.TOS.AvatarMatrixSoaks["Left"]),
            ["TaintedSoakRight"] = hcTk.GroupNameToGUIDs(hcTk.DB.profile.Assignments.TOS.AvatarMatrixSoaks["Right"]),
            ["type"] = "AvatarTaintedSoakGroups",
        }

        if (#data["TaintedSoakLeft"] > 0 or #data["TaintedSoakRight"] > 0) then
            hcTk.DebugPrint("Sending Tainted Soak Groups")
            hcTk:SendCommMessage("HCToolkitBoss", hcTk:Serialize(data), "RAID")
        end
    end
end

function hcTk.AvatarWipeTainted()
    HCToolkitGroupFrame:SetScript("OnUpdate", nil)
    hcTk.FrameUpdateText(HCToolkitGroupFrame, "")
    hcTk.FrameUpdateText(HCToolkitPersonalFrame, "")

    hcTk.Encounter["TaintedSoakLeft"] = nil
    hcTk.Encounter["TaintedSoakRight"] = nil
    for k, v in pairs(hcTk.Group) do
        v["TaintedExpiration"] = nil
        v["TaintedGroup"] = nil
        v["TaintedStacks"] = nil
        v["TaintedSoaking"] = nil
    end
end

function hcTk.AvatarAvalanchReminder()
    hcTk.FrameUpdateText(HCToolkitPersonalFrame, "Avalanche Potion")
    C_Timer.After(3, function() hcTk.FrameUpdateText(HCToolkitPersonalFrame, "") end)
end

hcTk.AvatarCombatLogFns = {
    ["SPELL_AURA_APPLIED"] = hcTk.AvatarSpellAuraApplied,
    ["SPELL_AURA_APPLIED_DOSE"] = hcTk.AvatarSpellAuraAppliedDose,
    ["SPELL_AURA_REMOVED"] = hcTk.AvatarSpellAuraRemoved,
    ["SPELL_CAST_SUCCESS"] = hcTk.AvatarSpellCastSuccess,
    ["UNIT_DIED"] = hcTk.AvatarUnitDied
}